from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.instrumentation.asyncio import AsyncioInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
)

from atropylib.telemetry.utils import get_resource


def init_trace_provider(service_name: str | None = None) -> None:
    resource = get_resource(service_name)
    provider = TracerProvider(resource=resource)
    span_exporter = OTLPSpanExporter("http://otel:4318/v1/traces")
    processor = BatchSpanProcessor(span_exporter)
    provider.add_span_processor(processor)

    # Sets the global default tracer provider
    trace.set_tracer_provider(provider)
    AsyncioInstrumentor().instrument()
    RequestsInstrumentor().instrument()



import datetime
import logging
from json import dumps
from typing import Any

from opentelemetry._logs import set_logger_provider
from opentelemetry.exporter.otlp.proto.http._log_exporter import (
    OTLPLogExporter,
)
from opentelemetry.instrumentation.logging import LoggingInstrumentor
from opentelemetry.sdk._logs import LoggerProvider, LogRecord
from opentelemetry.sdk._logs.export import (
    BatchLogRecordProcessor,
    ConsoleLogExporter,
    SimpleLogRecordProcessor,
)

from atropylib.errors import DeveloperError
from atropylib.telemetry.logs.line_fix_handler import LineFixLoggingHandler
from atropylib.telemetry.utils import get_resource

LOGGER_PROVIDER: LoggerProvider | None = None
APP_SHUTDOWN: bool = False
HANDLER: LineFixLoggingHandler | None = None


class StructuredLogger:
    def __init__(self, name: str, ctx: dict[str, Any] | None = None):
        self._logger = logging.getLogger(name)
        self._handler_added: bool = False
        self._ctx: dict[str, Any] = ctx or {}

    def _lazy_init(self):
        global LOGGER_PROVIDER

        if not LOGGER_PROVIDER:
            raise DeveloperError("Logger provider not initialized, but it should have been via init logger provider.")

        if not self._handler_added:
            add_otlp_handler(self._logger)
            self._handler_added = True

    def bind(self, **ctx: Any) -> "StructuredLogger":
        return StructuredLogger(self._logger.name, self._ctx | ctx)

    def debug(self, msg: str, **kwargs: Any) -> None:
        self._lazy_init()
        self._logger.debug(msg, extra=self._ctx | kwargs)

    def info(self, msg: str, **kwargs: Any) -> None:
        self._lazy_init()
        self._logger.handlers.clear()
        self._logger.info(msg, extra=self._ctx | kwargs)

    def warning(self, msg: str, **kwargs: Any) -> None:
        self._lazy_init()
        self._logger.warning(msg, extra=self._ctx | kwargs)

    def error(self, msg: str, **kwargs: Any) -> None:
        self._lazy_init()
        self._logger.error(msg, extra=self._ctx | kwargs)

    def critical(self, msg: str, **kwargs: Any) -> None:
        self._lazy_init()
        self._logger.critical(msg, extra=self._ctx | kwargs)


def ns_to_iso_str(nanoseconds: int | None) -> str:
    if nanoseconds is None:
        return ""

    """Get an ISO 8601 string from time_ns value."""
    ts = datetime.datetime.fromtimestamp(nanoseconds / 1e9, tz=datetime.timezone.utc)
    return ts.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def log_format(record: LogRecord) -> str:
    msg = {
        "body": record.body,
        "severity": record.severity_text,
        "timestamp": ns_to_iso_str(record.timestamp),
        "attributes": dict(record.attributes) if bool(record.attributes) else None,  # type: ignore[arg-type]
    }

    return dumps(msg, indent=4) + "\n"


def shutdown_logger_provider():
    global APP_SHUTDOWN
    APP_SHUTDOWN = True

    global LOGGER_PROVIDER
    if LOGGER_PROVIDER:
        LOGGER_PROVIDER.shutdown()
        LOGGER_PROVIDER = None

    global HANDLER
    if HANDLER:
        HANDLER.close()
        HANDLER = None

    loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]

    # Remove all handlers from each logger
    for logger in loggers:
        logger.handlers.clear()


def add_otlp_handler(logger: logging.Logger):
    global LOGGER_PROVIDER, HANDLER

    if not LOGGER_PROVIDER:
        raise DeveloperError("Logger provider not initialized, but it should have been via init logger provider.")
    if not HANDLER:
        raise DeveloperError("Handler not initialized, but it should have been via init logger provider.")

    logger.addHandler(HANDLER)
    logger.setLevel(logging.INFO)


def init_logger_provider(service_name: str | None = None):
    global LOGGER_PROVIDER, HANDLER

    if not LOGGER_PROVIDER:
        resource = get_resource(service_name)
        LOGGER_PROVIDER = LoggerProvider(resource=resource)
        set_logger_provider(LOGGER_PROVIDER)

        exporter_otlp = OTLPLogExporter(endpoint="http://otel:4318/v1/logs")
        exporter_console = ConsoleLogExporter(formatter=log_format)
        LOGGER_PROVIDER.add_log_record_processor(BatchLogRecordProcessor(exporter_otlp))
        LOGGER_PROVIDER.add_log_record_processor(SimpleLogRecordProcessor(exporter_console))

    if not HANDLER:
        root = logging.getLogger()
        root.handlers.clear()

        HANDLER = LineFixLoggingHandler(level=logging.INFO, logger_provider=LOGGER_PROVIDER)
        LoggingInstrumentor().instrument(set_logging_format=False)

        root.addHandler(HANDLER)


def init_test_logger_provider():
    global LOGGER_PROVIDER, HANDLER

    if not LOGGER_PROVIDER:
        LOGGER_PROVIDER = LoggerProvider()
        set_logger_provider(LOGGER_PROVIDER)

        exporter_console = ConsoleLogExporter(formatter=log_format)
        LOGGER_PROVIDER.add_log_record_processor(SimpleLogRecordProcessor(exporter_console))

    if not HANDLER:
        root = logging.getLogger()
        root.handlers.clear()

        HANDLER = LineFixLoggingHandler(level=logging.INFO, logger_provider=LOGGER_PROVIDER)
        LoggingInstrumentor().instrument(set_logging_format=False)

        root.addHandler(HANDLER)



import os

from opentelemetry.sdk.resources import SERVICE_NAME, Resource

from atropylib.errors import DeveloperError


def get_resource(service_name: str | None = None) -> Resource:
    service_name = service_name or os.getenv("ATRO_SERVICE_NAME", None)

    if not service_name:
        raise DeveloperError(
            "Service name must be provided either thorugh the function argument "
            "or the environment variable (ATRO_SERVICE_NAME)."
        )

    return Resource({SERVICE_NAME: service_name})



from opentelemetry import metrics
from opentelemetry.exporter.otlp.proto.http.metric_exporter import OTLPMetricExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader

from atropylib.telemetry.utils import get_resource


def init_meter_provider(service_name: str | None = None) -> None:
    resource = get_resource(service_name)
    reader = PeriodicExportingMetricReader(OTLPMetricExporter(endpoint="http://otel:4318/v1/metrics"))
    meterProvider = MeterProvider(resource=resource, metric_readers=[reader])
    metrics.set_meter_provider(meterProvider)



from typing import Any

LOGGING_CONFIG: dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "default": {
            "class": "atropylib.telemetry.logging.line_fix_handler.LineFixLoggingHandler",
        },
        "access": {
            "class": "atropylib.telemetry.logging.line_fix_handler.LineFixLoggingHandler",
        },
        "error": {
            "class": "atropylib.telemetry.logging.line_fix_handler.LineFixLoggingHandler",
        },
    },
    "loggers": {
        "uvicorn": {"handlers": ["default"], "level": "INFO", "propagate": False},
        "uvicorn.error": {"level": "INFO"},
        "uvicorn.access": {"handlers": ["access"], "level": "INFO", "propagate": False},
    },
}



import base64
import json
from datetime import datetime
from typing import Any

from nats.aio.msg import Msg
from pydantic import BaseModel


class Message(BaseModel):
    class Config:
        frozen = True

    source: str
    uuid: str
    time: datetime
    payload: dict[str, Any]
    extra_metadata: dict[str, str]
    subject: str | None = None

    def to_nats_msg(self) -> bytes:
        """
        Convert the message to a NATS message.
        """
        # Base64 encode the payload
        payload = base64.b64encode(json.dumps(self.payload).encode()).decode()
        metadata = {
            "source": self.source,
            "uuid": self.uuid,
            "time": self.time.isoformat(),
            **self.extra_metadata,
        }
        if self.subject:
            metadata["subject"] = self.subject
        uuid = self.uuid

        return json.dumps({"Metadata": metadata, "Payload": payload, "UUID": uuid}).encode("utf-8")

    @staticmethod
    def from_nats_msg(msg: Msg) -> "Message":
        """
        Convert a NATS message to a Message object.
        """
        data = msg.data.decode("utf-8")
        json_data: dict[str, Any] = json.loads(data)  # type: ignore[assignment]
        payload: dict[str, Any] = json.loads(base64.b64decode(json_data["Payload"]).decode())  # type: ignore[assignment]
        uuid: str = json_data["UUID"]
        msg_metadata: dict[str, str] = json_data["Metadata"]

        time = datetime.fromisoformat(msg_metadata["time"])
        source = msg_metadata["source"]
        subject = msg_metadata.get("subject", None)
        extra_metadata = {
            k: v
            for k, v in msg_metadata.items()
            if k
            not in [
                "source",
                "time",
                "subject",
            ]
        }

        return Message(
            source=source,
            uuid=uuid,
            time=time,
            payload=payload,
            extra_metadata=extra_metadata,
            subject=subject,
        )



import inspect
import logging

from opentelemetry.sdk._logs import LoggingHandler
from typing_extensions import override


class LineFixLoggingHandler(LoggingHandler):
    """
    A class to account for the fact that we use Logger class and need to look one stack higher to catch actual caller.
    """

    @override
    def emit(self, record: logging.LogRecord) -> None:
        """
        Emit a record. Skip emitting if logger is NoOp.

        The record is translated to OTel format, and then sent across the pipeline.
        """

        # WARN: It doens't work with asyncio stuff :(

        frame = inspect.stack()[-1]
        filename = frame.filename
        lineno = frame.lineno

        record.funcName = frame.function
        record.filename = filename
        record.pathname = filename
        record.lineno = lineno

        super().emit(record)



"""
Example usage

async def main():
    # Example usage
    nats_client = NATsClient()

    message = Message(
        source="example_source",
        uuid="example_uuid",
        time=datetime.now(),
        payload={"key": "value"},
        extra_metadata={"meta_key": "meta_value"},
        subject="test",
    )

    from time import time

    start = time()
    await nats_client.publish("test", message)
    end = time()
    print(f"Published message in {end - start} seconds")

    start = time()
    await nats_client.publish("test", message)
    end = time()
    print(f"Published message in {end - start} seconds")

    start = time()
    await nats_client.publish("test", message)
    end = time()
    print(f"Published message in {end - start} seconds")

    start = time()
    await nats_client.publish("blah", message)
    end = time()
    print(f"Published message in {end - start} seconds")

    start = time()
    await nats_client.publish("blah", message)
    end = time()
    print(f"Published message in {end - start} seconds")

    start = time()
    await nats_client.publish("test", message)
    end = time()
    print(f"Published message in {end - start} seconds")

    print("Waiting for message...")
    counter = 0
    async for i in nats_client.subscribe("test"):
        print(i)
        counter += 1
        print(counter)


asyncio.run(main())
"""

import asyncio
import os
from functools import cached_property
from typing import Any, AsyncGenerator

from nats import connect
from nats.aio.client import Client as NATS
from nats.js import JetStreamContext
from nats.js.api import AckPolicy, ConsumerConfig, DeliverPolicy
from pydantic import BaseModel, PrivateAttr

from atropylib.pubsub.message import Message

_SVC_NAME = os.getenv("ATRO_SVC_NAME", "")
_DEFAULT_CONSUMER_CONFIG = ConsumerConfig(
    ack_policy=AckPolicy.EXPLICIT,
    flow_control=True,
    durable_name=_SVC_NAME,
    idle_heartbeat=30,
)


async def get_nats_client(nats_url: str | None = None) -> NATS:
    """
    Connect to a NATS server and return the JetStream context.

    If `url` is not provided, we check the environment variable `ATRO_NATS_URL` for the server URL.
    """
    url = nats_url or os.getenv("ATRO_NATS_URL", None)

    if url is None:
        raise ValueError("NATs url is required but it was not provided and neither was env ATRO_NATS_URL.")

    # Connect to the NATS server
    nc = await connect(url)

    return nc


async def get_jetstream_client(nats_url: str | None = None) -> JetStreamContext:
    """
    Connect to a NATS server and return the JetStream context.

    If `url` is not provided, we check the environment variable `ATRO_NATS_URL` for the server URL.
    """
    nc = await get_nats_client(nats_url)

    return nc.jetstream()


class NATSClient(BaseModel):
    """
    A client for interacting with NATS JetStream.
    """

    nats_url: str | None = None
    _added_streams: set[str] = PrivateAttr(default_factory=set)

    @cached_property
    def _nc(self) -> NATS:
        """
        Return the NATS client.
        """
        import nest_asyncio

        nest_asyncio.apply()
        coro = get_nats_client(self.nats_url)
        nc = asyncio.run(coro)
        return nc

    @cached_property
    def _js(self) -> JetStreamContext:
        """
        Return the JetStream context.
        """
        return self._nc.jetstream()

    async def publish(self, subject: str, message: Message) -> None:
        """
        Publish a message to a subject.
        """
        if subject not in self._added_streams:
            await self._js.add_stream(name=subject, subjects=[subject])
            self._added_streams.add(subject)

        # Using jetstream publish is flaky so using nc.request directly.
        # Can use nc with request to impose a timeout but that is never been needed and imposes a huge
        # slow down so using .publish instaed with nats client.
        await self._nc.publish(
            subject=subject,
            payload=message.to_nats_msg(),
            headers=None,
        )

    async def subscribe(
        self,
        subject: str,
        deliver_policy: DeliverPolicy = DeliverPolicy.NEW,
    ) -> AsyncGenerator[Message, Any]:
        global _DEFAULT_CONSUMER_CONFIG, _SVC_NAME
        sub = await self._js.subscribe(
            subject=subject,
            durable=_SVC_NAME,
            config=_DEFAULT_CONSUMER_CONFIG,
            manual_ack=True,
            flow_control=True,
            deliver_policy=deliver_policy,
        )
        try:
            async for msg in sub.messages:
                # Order here matters, if we can't decode the message, we don't want to ack it
                message = Message.from_nats_msg(msg)
                await msg.ack()

                yield message
        finally:
            await sub.unsubscribe()
